import random
import character
import config

ENEMY_WALK_SPEEDS = [
  50,
  70,
  90,
  110,
  130
]

ENEMY_RUN_SPEEDS = [
  110,
  130,
  150,
  170,
  190,
]

class Lazy_Time_Accumulator:
  def __init__(self, time_to_accumulate):
    self.time_to_accumulate = time_to_accumulate + (random.random() * 2)
    self.accumulated_time = 0
  def accumulate(self, delta):
    self.accumulated_time += delta
  def time_is_up(self):
    return self.accumulated_time > self.time_to_accumulate
  def reset(self):
    self.accumulated_time = 0

class Behaviour(object):
  def __init__(self, parent, time):
    self.parent = parent
    self.timer = Lazy_Time_Accumulator(time)
    self.last_state_change = None
  def tick(self, time_delta):
    self.timer.accumulate(time_delta)
    self.signed_z_cached_delta = self.parent.world.player.pos.z - self.parent.pos.z
    self.signed_cached_delta = self.parent.world.player.pos.x - self.parent.pos.x
    self.cached_delta = abs(self.parent.world.player.pos.x - self.parent.pos.x)
    if self.last_state_change and not self.last_state_change.changed:
      self.last_state_change.do_action()
      return

class Consider_behaviour_change_mixin:
  def consider_change(self):
    if self.cached_delta < 400:
      self.parent.behaviour = Persuit(self.parent)
    if not self.timer.time_is_up():
      return
    self.parent.release()
    choice = random.choice(self.behaviour_choices)
    self.parent.behaviour = choice(self.parent)

class Get_Distracted_Mixin:
  def become_distracted(self):
    if not self.timer.time_is_up():
      return
    if self.cached_delta < 400:
      return
    self.parent.release()
    if random.random() > 0.5:
      self.parent.behaviour = Loiter(self.parent)
    else:
      self.parent.behaviour = Meander(self.parent)

class Player_Attack_Mixin:
  def move_towards_player(self):
    delta_x = self.parent.world.player.pos.x - self.parent.pos.x
    if self.parent.facing == character.FACING_LEFT:
      if delta_x < 0:
        self.last_state_change = Move_left_from_facing_left(self.parent)
        """
        config.pr(config.print_enemy_behaviour, "facing left, move in direction")
        self.last_state_change = self.parent.move_direction
        self.parent.move_direction()
        self.parent.facing = character.FACING_LEFT
        """
      else:
        self.last_state_change = Move_right_from_facing_left(self.parent)
        """
        config.pr(config.print_enemy_behaviour, "facing left, move opposite")
        self.parent.move_opposite()
        self.parent.facing = character.FACING_RIGHT
        """
    else:
      if delta_x < 0:
        self.last_state_change = Move_left_from_facing_right(self.parent)
        """
        config.pr(config.print_enemy_behaviour, "facing right, move opposite")
        self.parent.move_opposite()
        self.parent.facing = character.FACING_LEFT
        """
      else:
        self.last_state_change = Move_right_from_facing_right(self.parent)
        """
        config.pr(config.print_enemy_behaviour, "facing right, move in direction")
        self.parent.move_direction()
        self.parent.facing = character.FACING_RIGHT
        """
        
  def is_enemy_behind_me(self):
    if self.parent.facing == character.FACING_LEFT:
      return self.signed_cached_delta > 0
    if self.parent.facing == character.FACING_RIGHT:
      return self.signed_cached_delta < 0

  def is_enemy_far_away(self):
    if self.cached_delta > 50 or abs(self.signed_z_cached_delta) > 4:
      return True

class Loiter(Behaviour, Consider_behaviour_change_mixin):
  def __init__(self, parent):
    config.pr(config.print_enemy_behaviour, "Loitering")
    super(Loiter, self).__init__(parent, 2)
    self.behaviour_choices = [Meander, Persuit]
  def tick(self, delta):
    super(Loiter, self).tick(delta)
    self.consider_change() 

class Meander(Behaviour, Consider_behaviour_change_mixin, Player_Attack_Mixin):
  def __init__(self, parent):
    super(Meander, self).__init__(parent, 2)
    self.behaviour_choices = [Loiter, Persuit]
    self.parent.set_max_speed(random.choice(ENEMY_WALK_SPEEDS))
    self.move_towards_player()
    config.pr(config.print_enemy_behaviour, "Meandering")
  def tick(self, delta):
    super(Meander, self).tick(delta)
    self.consider_change()

class Cachable_state_transitions(object):
  def __init__(self, parent, changed = False):
    self.parent = parent
    self.changed = changed 

class Face_left(Cachable_state_transitions):
  def __init__(self, parent):
    super(Face_left, self).__init__(parent)
  def do_action(self):
    config.pr(config.print_enemy_behaviour, "facing left")
    self.parent.face_left()

class Face_right(Cachable_state_transitions):
  def __init__(self, parent):
    super(Face_right, self).__init__(parent)
  def do_action(self):
    config.pr(config.print_enemy_behaviour, "facing right")
    self.parent.face_right()

class Move_left_from_facing_left(Cachable_state_transitions):
  def __init__(self, parent):
    super(Move_left_from_facing_left, self).__init__(parent)
  def do_action(self):
    self.changed = self.parent.move_direction()
    self.parent.face_left()
    
class Move_right_from_facing_left(Cachable_state_transitions):
  def __init__(self, parent):
    super(Move_right_from_facing_left, self).__init__(parent)
  def do_action(self):
    self.changed = self.parent.move_opposite()
    self.parent.face_right()

class Move_left_from_facing_right(Cachable_state_transitions):
  def __init__(self, parent):
    super(Move_left_from_facing_right, self).__init__(parent)
  def do_action(self):
    self.changed = self.parent.move_opposite()
    self.parent.face_left()
    
class Move_right_from_facing_right(Cachable_state_transitions):
  def __init__(self, parent):
    super(Move_right_from_facing_right, self).__init__(parent)
  def do_action(self):
    self.changed = self.parent.move_direction()
    self.parent.face_right()

class Persuit(Behaviour, Get_Distracted_Mixin, Player_Attack_Mixin):
  def __init__(self, parent):
    super(Persuit, self).__init__(parent, 5)
    config.pr(config.print_enemy_behaviour, "Persuing")
    self.parent.set_max_speed(random.choice(ENEMY_RUN_SPEEDS))
    self.move_towards_player()
    self.moving_x = True
  def tick(self, delta):
    super(Persuit, self).tick(delta)
    self.become_distracted()
    z_delta = self.parent.world.player.pos.z - self.parent.pos.z 
    abs_z_delta = abs(z_delta)
    if abs_z_delta > 2:
      self.parent.pos.z += (z_delta / abs_z_delta) * 20* delta
    if self.moving_x and not self.is_enemy_far_away():
      config.pr(config.print_enemy_behaviour, "going out of move")
      self.moving_x = False
      self.parent.release()
    if self.is_enemy_far_away():
      #config.pr(config.print_enemy_behaviour, "going into move from stationary")
      self.moving_x = True
      self.move_towards_player()
      return
    if abs_z_delta < 2:
      config.pr(config.print_enemy_behaviour, "going into an  attack state")
      self.parent.release()
      self.parent.behaviour = Attack(self.parent)

class Attack(Behaviour, Get_Distracted_Mixin, Player_Attack_Mixin):
  def __init__(self, parent):
    super(Attack, self).__init__(parent, 2)
    self.lazy_attack = Lazy_Time_Accumulator(0)
    config.pr(config.print_enemy_behaviour, "Attacking")
  def tick(self, delta):
    super(Attack, self).tick(delta)
    if self.is_enemy_far_away():
      config.pr(config.print_enemy_behaviour, "going into persuit")
      self.parent.behaviour = Persuit(self.parent)
    if self.is_enemy_behind_me():
      config.pr(config.print_enemy_behaviour, "going in to turn round")
      if self.parent.facing == character.FACING_LEFT:
        self.last_state_change = Face_right(self.parent)
      else:
        self.last_state_change = Face_left(self.parent)

    if not self.parent.state.is_trying_attack():
      self.lazy_attack.accumulate(delta)
      self.parent.high_attack()
    self.become_distracted()

class Wait_for_recovery(Behaviour, Player_Attack_Mixin):
  def __init__(self, parent):
    super(Wait_for_recovery, self).__init__(parent, 5)
    self.lazy_attack = Lazy_Time_Accumulator(0)
    config.pr(config.print_enemy_behaviour, "WaitingForRecovery")
  def tick(self, delta):
    super(Wait_for_recovery, self).tick(delta)
    if not self.parent.world.player.state.is_safe():
      self.parent.behaviour = Persuit(self.parent)
