import random
from character import FACING_LEFT, FACING_RIGHT
from enemy import Enemy
from vector import Vector3

DIRECTIONS = {
  "Left": FACING_LEFT,
  "Right": FACING_RIGHT,
}

INIT_POSITIONS = {
  FACING_LEFT: (1000, 0),
  FACING_RIGHT: (-200, 0),
}

ENEMIES = {
  "Grunt": Enemy,
}

class Wait_node:
  def __init__(self, fight_client):
    self.fight_client = fight_client
  def work(self, time_delta):
    if len(self.fight_client.objects) == 1:
      return True
    return False
  def __str__(self):
    return "Wait"
  def reset(self):
    pass

class Sleep_node:
  def __init__(self, duration):
    self.wait = duration
    self.time = 0
  def work(self, time_delta):
    self.time += time_delta
    return self.time > self.wait
  def reset(self):
    self.time = 0

class Message_node:
  def __init__(self, fight_client, message, duration, centred = False):
    self.fight_client = fight_client
    if duration < 0:
      self.wait = 10000
      self.forever = True
    else:
      self.wait = duration
      self.forever = False
    self.time = 0
    self.message = message
    self.centred = centred
    self.shown_message = False
  def work(self, time_delta):
    if not self.shown_message:
      if self.centred:
        self.fight_client.set_message (self.message,self.wait)
      else:
        self.fight_client.set_multiline_message (self.message,self.wait)
      self.shown_message = True
    self.time += time_delta
    return self.forever or self.time > self.wait
  def reset(self):
    self.time = 0
    self.shown_message = False



class Spawn_node:
  def __init__(self, fight_client, delta, direction, z_offset, EnemyObj):
    self.fight_client = fight_client
    self.wait = delta
    self.time = 0
    self.direction = direction
    self.z_offset = z_offset
    self.EnemyObj = EnemyObj
  def work(self, time_delta):
    self.time += time_delta
    x, y = INIT_POSITIONS[self.direction]
    x += self.fight_client.screen_pos.x
    if x < 200:
      x = 1000
    if x > 2200:
      x = 1000
    if self.time > self.wait:
      self.enemy = self.EnemyObj(self.fight_client, Vector3(x, self.fight_client.screen_pos.y + y, self.z_offset), self.direction)
      self.fight_client.add_character(self.enemy)
      return True
    return False
  def __str__(self):
    return "Spawn"
  def reset(self):
    self.time = 0

class End_node:
  def __init__(self, fight_client):
    self.fight_client = fight_client
  def work(self, time_delta):
    # Trigger end of level, done in client
    return False
  def __str__(self):
    return "End"
  def reset(self):
    pass

class Director:
  def __init__(self, fight_client, datafile = "data/director_script"):
    self.events = []
    self.index = 0
    message_centred = False
    message_duration = 0
    for line in open(datafile):
      line = line.strip()
      if message_duration:
        #a message was found last time
        self.events.append(Message_node(fight_client,line,message_duration,message_centred))
        message_duration = 0
        continue
      if line == "":
        continue
      if line.startswith("#"):
        continue
      if line == "Wait":
        self.events.append(Wait_node(fight_client))
      elif line.startswith("Add"):
        delta, direction, y_offset, obj_class = line.split()[1:]
        self.events.append(Spawn_node(fight_client, float(delta), DIRECTIONS[direction], float(y_offset), ENEMIES[obj_class]))
      elif line.startswith("Message"):
        message_duration = float(line.split()[1])
        message_centred = False
      elif line.startswith("CentreMessage"):
        message_duration = float(line.split()[1])
        message_centred = True
      elif line.startswith("Sleep"):
        duration = line.split()[1]
        self.events.append(Sleep_node(float(duration)))

        
    self.events.append(Wait_node(fight_client))
    self.events.append(End_node(fight_client))
    self.current = self.events[self.index]
  
  def done(self):
    return self.index == (len(self.events) - 1)
    
  def restart(self):
    self.index = 0
    for event in self.events:
      event.reset()
    self.current = self.events[self.index]
    
  def tick(self, time_delta):
    # changed the default position of 0, 0, to be from the right or left
    finished = self.current.work(time_delta)
    if finished:
      self.index += 1
      self.current = self.events[self.index]
      
